import datetime
import os
import pathlib
from decimal import Decimal

import pytest
from arro3.core import Array, DataType, Table
from arro3.core import Field as ArrowField

from deltalake import (
    CommitProperties,
    DeltaTable,
    Field,
    Schema,
    write_deltalake,
)
from deltalake.query import QueryBuilder
from deltalake.schema import PrimitiveType


@pytest.mark.parametrize("streaming", (True, False))
def test_merge_when_matched_delete_wo_predicate(
    tmp_path: pathlib.Path, sample_table: Table, streaming: bool
):
    write_deltalake(tmp_path, sample_table, mode="append")

    dt = DeltaTable(tmp_path)

    source_table = Table(
        {
            "id": Array(["5"], ArrowField("id", DataType.string_view(), nullable=True)),
            "weight": Array([105], ArrowField("id", DataType.int32(), nullable=True)),
        }
    )

    commit_properties = CommitProperties(custom_metadata={"userName": "John Doe"})
    dt.merge(
        source=source_table,
        predicate="t.id = s.id",
        source_alias="s",
        target_alias="t",
        commit_properties=commit_properties,
        streamed_exec=streaming,
    ).when_matched_delete().execute()

    nrows = 4

    expected = Table(
        {
            "id": Array(
                ["1", "2", "3", "4"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                list(range(nrows)),
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                list(range(nrows)),
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [False] * nrows,
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
        },
    )

    result = (
        QueryBuilder()
        .register("tbl", dt)
        .execute("select * from tbl order by id asc")
        .read_all()
    )
    last_action = dt.history(1)[0]

    assert last_action["operation"] == "MERGE"
    assert last_action["userName"] == "John Doe"
    assert result == expected


@pytest.mark.parametrize("streaming", (True, False))
def test_merge_when_matched_delete_with_predicate(
    tmp_path: pathlib.Path, sample_table: Table, streaming: bool
):
    write_deltalake(tmp_path, sample_table, mode="append")

    dt = DeltaTable(tmp_path)

    source_table = Table(
        {
            "id": Array(
                ["5", "4"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "weight": Array(
                [1, 2],
                ArrowField("weight", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [1, 2],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [True, False],
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
            "customer": Array(
                ["Adam", "Patrick"],
                ArrowField("customer", type=DataType.string_view(), nullable=True),
            ),
        },
    )

    dt.merge(
        source=source_table,
        predicate="t.id = s.id",
        source_alias="s",
        target_alias="t",
        streamed_exec=streaming,
    ).when_matched_delete("s.deleted = True").execute()

    nrows = 4
    expected = Table(
        {
            "id": Array(
                ["1", "2", "3", "4"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                list(range(nrows)),
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                list(range(nrows)),
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [False] * nrows,
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
        },
    )

    result = (
        QueryBuilder()
        .register("tbl", dt)
        .execute("select * from tbl order by id asc")
        .read_all()
    )
    last_action = dt.history(1)[0]

    assert last_action["operation"] == "MERGE"
    assert result == expected


@pytest.mark.parametrize("streaming", (True, False))
def test_merge_when_matched_update_wo_predicate(
    tmp_path: pathlib.Path, sample_table: Table, streaming: bool
):
    write_deltalake(tmp_path, sample_table, mode="append")

    dt = DeltaTable(tmp_path)

    source_table = Table(
        {
            "id": Array(
                ["4", "5"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [10, 100],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [10, 20],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
        },
    )

    dt.merge(
        source=source_table,
        predicate="t.id = s.id",
        source_alias="s",
        target_alias="t",
        streamed_exec=streaming,
    ).when_matched_update({"price": "s.price", "sold": "s.sold"}).execute()

    expected = Table(
        {
            "id": Array(
                ["1", "2", "3", "4", "5"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [0, 1, 2, 10, 100],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [0, 1, 2, 10, 20],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [False] * 5,
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
        },
    )

    result = (
        QueryBuilder()
        .register("tbl", dt)
        .execute("select * from tbl order by id asc")
        .read_all()
    )

    last_action = dt.history(1)[0]

    assert last_action["operation"] == "MERGE"
    assert result == expected


def test_merge_when_matched_update_wo_predicate_with_schema_evolution(
    tmp_path: pathlib.Path, sample_table: Table
):
    write_deltalake(tmp_path, sample_table, mode="append")

    dt = DeltaTable(tmp_path)

    source_table = Table(
        {
            "id": Array(
                ["4", "5"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [10, 100],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [10, 20],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "customer": Array(
                ["john", "doe"],
                ArrowField("customer", type=DataType.string_view(), nullable=True),
            ),
        },
    )

    dt.merge(
        source=source_table,
        predicate="t.id = s.id",
        source_alias="s",
        target_alias="t",
        merge_schema=True,
    ).when_matched_update(
        {"price": "s.price", "sold": "s.sold+int'10'", "customer": "s.customer"}
    ).execute()

    expected = Table(
        {
            "id": Array(
                ["1", "2", "3", "4", "5"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [0, 1, 2, 10, 100],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [0, 1, 2, 20, 30],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [False] * 5,
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
            "customer": Array(
                [None, None, None, "john", "doe"],
                ArrowField("customer", type=DataType.string_view(), nullable=True),
            ),
        },
    )

    result = (
        QueryBuilder()
        .register("tbl", dt)
        .execute("select * from tbl order by id asc")
        .read_all()
    )

    last_action = dt.history(1)[0]

    assert last_action["operation"] == "MERGE"
    assert result.schema == expected.schema
    assert result == expected


def test_merge_when_matched_update_wo_predicate_and_insert_with_schema_evolution(
    tmp_path: pathlib.Path, sample_table: Table
):
    write_deltalake(tmp_path, sample_table, mode="append")

    dt = DeltaTable(tmp_path)

    source_table = Table(
        {
            "id": Array(
                ["4", "5"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [10, 100],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [10, 20],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "customer": Array(
                ["john", "doe"],
                ArrowField("customer", type=DataType.string_view(), nullable=True),
            ),
        },
    )

    dt.merge(
        source=source_table,
        predicate="t.id = s.id",
        source_alias="s",
        target_alias="t",
        merge_schema=True,
    ).when_matched_update(
        {"price": "s.price", "sold": "s.sold+int'10'", "customer": "s.customer"}
    ).when_not_matched_insert_all().execute()

    expected = Table(
        {
            "id": Array(
                ["1", "2", "3", "4", "5"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [0, 1, 2, 10, 100],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [0, 1, 2, 20, 30],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [False] * 5,
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
            "customer": Array(
                [None, None, None, "john", "doe"],
                ArrowField("customer", type=DataType.string_view(), nullable=True),
            ),
        },
    )

    result = (
        QueryBuilder()
        .register("tbl", dt)
        .execute("select * from tbl order by id asc")
        .read_all()
    )

    last_action = dt.history(1)[0]

    assert last_action["operation"] == "MERGE"
    assert result.schema == expected.schema
    assert result == expected


@pytest.mark.parametrize("streaming", (True, False))
def test_merge_when_matched_update_all_wo_predicate(
    tmp_path: pathlib.Path, sample_table: Table, streaming: bool
):
    write_deltalake(tmp_path, sample_table, mode="append")

    dt = DeltaTable(tmp_path)

    source_table = Table(
        {
            "id": Array(
                ["4", "5"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [10, 100],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [10, 20],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [True, True],
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
            "weight": Array(
                [10, 15],
                ArrowField("weight", type=DataType.int64(), nullable=True),
            ),
        },
    )

    dt.merge(
        source=source_table,
        predicate="t.id = s.id",
        source_alias="s",
        target_alias="t",
        streamed_exec=streaming,
    ).when_matched_update_all().execute()

    expected = Table(
        {
            "id": Array(
                ["1", "2", "3", "4", "5"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [0, 1, 2, 10, 100],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [0, 1, 2, 10, 20],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [False, False, False, True, True],
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
        },
    )

    result = (
        QueryBuilder()
        .register("tbl", dt)
        .execute("select * from tbl order by id asc")
        .read_all()
    )
    last_action = dt.history(1)[0]

    assert last_action["operation"] == "MERGE"
    assert result == expected


@pytest.mark.parametrize("streaming", (True, False))
def test_merge_when_matched_update_all_with_exclude(
    tmp_path: pathlib.Path, sample_table: Table, streaming: bool
):
    write_deltalake(tmp_path, sample_table, mode="append")

    dt = DeltaTable(tmp_path)

    source_table = Table(
        {
            "id": Array(
                ["4", "5"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [10, 100],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [15, 25],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [True, True],
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
            "weight": Array(
                [10, 15],
                ArrowField("weight", type=DataType.int64(), nullable=True),
            ),
        }
    )

    dt.merge(
        source=source_table,
        predicate="t.id = s.id",
        source_alias="s",
        target_alias="t",
        streamed_exec=streaming,
    ).when_matched_update_all(except_cols=["sold"]).execute()

    expected = Table(
        {
            "id": Array(
                ["1", "2", "3", "4", "5"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [0, 1, 2, 10, 100],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [0, 1, 2, 3, 4],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [False, False, False, True, True],
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
        }
    )
    result = (
        QueryBuilder()
        .register("tbl", dt)
        .execute("select * from tbl order by id asc")
        .read_all()
    )
    last_action = dt.history(1)[0]

    assert last_action["operation"] == "MERGE"
    assert result == expected


@pytest.mark.parametrize("streaming", (True, False))
def test_merge_when_matched_update_with_predicate(
    tmp_path: pathlib.Path, sample_table: Table, streaming: bool
):
    write_deltalake(tmp_path, sample_table, mode="append")

    dt = DeltaTable(tmp_path)

    source_table = Table(
        {
            "id": Array(
                ["4", "5"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [10, 100],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [10, 20],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [False, True],
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
        }
    )

    dt.merge(
        source=source_table,
        source_alias="source",
        target_alias="target",
        predicate="target.id = source.id",
        streamed_exec=streaming,
    ).when_matched_update(
        updates={"price": "source.price", "sold": "source.sold"},
        predicate="source.deleted = False",
    ).execute()

    expected = Table(
        {
            "id": Array(
                ["1", "2", "3", "4", "5"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [0, 1, 2, 10, 4],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [0, 1, 2, 10, 4],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [False] * 5,
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
        }
    )
    result = (
        QueryBuilder()
        .register("tbl", dt)
        .execute("select * from tbl order by id asc")
        .read_all()
    )
    last_action = dt.history(1)[0]

    assert last_action["operation"] == "MERGE"
    assert result == expected


@pytest.mark.parametrize("streaming", (True, False))
def test_merge_when_not_matched_insert_wo_predicate(
    tmp_path: pathlib.Path, sample_table: Table, streaming: bool
):
    write_deltalake(tmp_path, sample_table, mode="append")

    dt = DeltaTable(tmp_path)

    source_table = Table(
        {
            "id": Array(
                ["4", "6"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [10, 100],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [10, 20],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [False, False],
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
        }
    )

    dt.merge(
        source=source_table,
        source_alias="source",
        target_alias="target",
        predicate="target.id = source.id",
        streamed_exec=streaming,
    ).when_not_matched_insert(
        updates={
            "id": "source.id",
            "price": "source.price",
            "sold": "source.sold",
            "deleted": "False",
        }
    ).execute()

    expected = Table(
        {
            "id": Array(
                ["1", "2", "3", "4", "5", "6"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [0, 1, 2, 3, 4, 100],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [0, 1, 2, 3, 4, 20],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [False] * 6,
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
        }
    )

    result = (
        QueryBuilder()
        .register("tbl", dt)
        .execute("select * from tbl order by id asc")
        .read_all()
    )
    last_action = dt.history(1)[0]

    assert last_action["operation"] == "MERGE"
    assert result == expected


@pytest.mark.parametrize("streaming", (True, False))
def test_merge_when_not_matched_insert_with_predicate(
    tmp_path: pathlib.Path, sample_table: Table, streaming: bool
):
    write_deltalake(tmp_path, sample_table, mode="append")

    dt = DeltaTable(tmp_path)

    source_table = Table(
        {
            "id": Array(
                ["6", "10"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [10, 100],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [10, 20],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [False, False],
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
        }
    )

    dt.merge(
        source=source_table,
        source_alias="source",
        target_alias="target",
        predicate="target.id = source.id",
        streamed_exec=streaming,
    ).when_not_matched_insert(
        updates={
            "id": "source.id",
            "price": "source.price",
            "sold": "source.sold",
            "deleted": "False",
        },
        predicate="source.price < 50",
    ).execute()

    expected = Table(
        {
            "id": Array(
                ["1", "2", "3", "4", "5", "6"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [0, 1, 2, 3, 4, 10],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [0, 1, 2, 3, 4, 10],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [False] * 6,
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
        }
    )

    result = (
        QueryBuilder()
        .register("tbl", dt)
        .execute("select * from tbl order by id asc")
        .read_all()
    )
    last_action = dt.history(1)[0]

    assert last_action["operation"] == "MERGE"
    assert result == expected


def test_merge_when_not_matched_insert_with_predicate_schema_evolution(
    tmp_path: pathlib.Path, sample_table: Table
):
    write_deltalake(tmp_path, sample_table, mode="append")

    dt = DeltaTable(tmp_path)

    source_table = Table(
        {
            "id": Array(
                ["6", "10"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [10, 100],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [10, 20],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "customer": Array(
                ["john", "doe"],
                ArrowField("customer", type=DataType.string_view(), nullable=True),
            ),
            "deleted": Array(
                [False, False],
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
        }
    )

    dt.merge(
        source=source_table,
        source_alias="source",
        target_alias="target",
        merge_schema=True,
        predicate="target.id = source.id",
    ).when_not_matched_insert(
        updates={
            "id": "source.id",
            "price": "source.price",
            "sold": "source.sold",
            "customer": "source.customer",
            "deleted": "False",
        },
        predicate="source.price < 50",
    ).execute()

    expected = Table(
        {
            "id": Array(
                ["1", "2", "3", "4", "5", "6"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [0, 1, 2, 3, 4, 10],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [0, 1, 2, 3, 4, 10],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [False] * 6,
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
            "customer": Array(
                [None, None, None, None, None, "john"],
                ArrowField("customer", type=DataType.string_view(), nullable=True),
            ),
        }
    )

    result = (
        QueryBuilder()
        .register("tbl", dt)
        .execute("select * from tbl order by id asc")
        .read_all()
    )
    last_action = dt.history(1)[0]

    assert last_action["operation"] == "MERGE"
    assert result.schema == expected.schema
    assert result == expected


@pytest.mark.parametrize("streaming", (True, False))
def test_merge_when_not_matched_insert_all_with_predicate(
    tmp_path: pathlib.Path, sample_table: Table, streaming: bool
):
    write_deltalake(tmp_path, sample_table, mode="append")

    dt = DeltaTable(tmp_path)

    source_table = Table(
        {
            "id": Array(
                ["6", "10"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [10, 100],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [10, 20],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [None, None],
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
        }
    )

    dt.merge(
        source=source_table,
        source_alias="source",
        target_alias="target",
        predicate="target.id = source.id",
        streamed_exec=streaming,
    ).when_not_matched_insert_all(
        predicate="source.price < 50",
    ).execute()

    expected = Table(
        {
            "id": Array(
                ["1", "2", "3", "4", "5", "6"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [0, 1, 2, 3, 4, 10],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [0, 1, 2, 3, 4, 10],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [False, False, False, False, False, None],
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
        }
    )

    result = (
        QueryBuilder()
        .register("tbl", dt)
        .execute("select * from tbl order by id asc")
        .read_all()
    )
    last_action = dt.history(1)[0]

    assert last_action["operation"] == "MERGE"
    assert result == expected


@pytest.mark.parametrize("streaming", (True, False))
def test_merge_when_not_matched_insert_all_with_exclude(
    tmp_path: pathlib.Path, sample_table: Table, streaming: bool
):
    write_deltalake(tmp_path, sample_table, mode="append")

    dt = DeltaTable(tmp_path)

    source_table = Table(
        {
            "id": Array(
                ["6", "9"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [10, 100],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [10, 20],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [None, None],
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
        }
    )

    dt.merge(
        source=source_table,
        source_alias="source",
        target_alias="target",
        predicate="target.id = source.id",
        streamed_exec=streaming,
    ).when_not_matched_insert_all(except_cols=["sold"]).execute()

    expected = Table(
        {
            "id": Array(
                ["1", "2", "3", "4", "5", "6", "9"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [0, 1, 2, 3, 4, 10, 100],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [0, 1, 2, 3, 4, None, None],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [False, False, False, False, False, None, None],
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
        }
    )

    result = (
        QueryBuilder()
        .register("tbl", dt)
        .execute("select * from tbl order by id asc")
        .read_all()
    )
    last_action = dt.history(1)[0]

    assert last_action["operation"] == "MERGE"
    assert result == expected


def test_merge_when_not_matched_insert_all_with_exclude_and_with_schema_evo(
    tmp_path: pathlib.Path, sample_table: Table
):
    write_deltalake(tmp_path, sample_table, mode="append")

    dt = DeltaTable(tmp_path)

    source_table = Table(
        {
            "id": Array(
                ["6", "9"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [10, 100],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [10, 20],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [None, None],
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
            "customer": Array(
                ["john", "doe"],
                ArrowField("customer", type=DataType.string_view(), nullable=True),
            ),
        }
    )

    dt.merge(
        source=source_table,
        source_alias="source",
        target_alias="target",
        merge_schema=True,
        predicate="target.id = source.id",
    ).when_not_matched_insert_all(except_cols=["sold"]).execute()

    expected = Table(
        {
            "id": Array(
                ["1", "2", "3", "4", "5", "6", "9"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [0, 1, 2, 3, 4, 10, 100],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [0, 1, 2, 3, 4, None, None],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [False, False, False, False, False, None, None],
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
            "customer": Array(
                [None, None, None, None, None, "john", "doe"],
                ArrowField("customer", type=DataType.string_view(), nullable=True),
            ),
        }
    )

    result = (
        QueryBuilder()
        .register("tbl", dt)
        .execute("select * from tbl order by id asc")
        .read_all()
    )
    last_action = dt.history(1)[0]

    assert last_action["operation"] == "MERGE"
    assert result.schema == expected.schema
    assert result == expected


@pytest.mark.parametrize("streaming", (True, False))
def test_merge_when_not_matched_insert_all_with_predicate_special_column_names(
    tmp_path: pathlib.Path, sample_table_with_spaces_numbers: Table, streaming: bool
):
    write_deltalake(tmp_path, sample_table_with_spaces_numbers, mode="append")

    dt = DeltaTable(tmp_path)

    source_table = Table(
        {
            "1id": Array(
                ["6", "10"],
                ArrowField("1id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [10, 100],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold items": Array(
                [10, 20],
                ArrowField("sold items", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [None, None],
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
        }
    )

    dt.merge(
        source=source_table,
        source_alias="source",
        target_alias="target",
        predicate="target.`1id` = source.`1id`",
        streamed_exec=streaming,
    ).when_not_matched_insert_all(
        predicate="source.price < 50",
    ).execute()

    expected = Table(
        {
            "1id": Array(
                ["1", "2", "3", "4", "5", "6"],
                ArrowField("1id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [0, 1, 2, 3, 4, 10],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold items": Array(
                [0, 1, 2, 3, 4, 10],
                ArrowField("sold items", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [False, False, False, False, False, None],
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
        }
    )

    result = (
        QueryBuilder()
        .register("tbl", dt)
        .execute("select * from tbl order by `1id` asc")
        .read_all()
    )
    last_action = dt.history(1)[0]

    assert last_action["operation"] == "MERGE"
    assert result == expected


@pytest.mark.parametrize("streaming", (True, False))
def test_merge_when_not_matched_by_source_update_wo_predicate(
    tmp_path: pathlib.Path, sample_table: Table, streaming: bool
):
    write_deltalake(tmp_path, sample_table, mode="append")

    dt = DeltaTable(tmp_path)

    source_table = Table(
        {
            "id": Array(
                ["6", "7"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [10, 100],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [10, 20],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [False, False],
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
        }
    )

    dt.merge(
        source=source_table,
        source_alias="source",
        target_alias="target",
        predicate="target.id = source.id",
        streamed_exec=streaming,
    ).when_not_matched_by_source_update(
        updates={
            "sold": "int'10'",
        }
    ).execute()

    expected = Table(
        {
            "id": Array(
                ["1", "2", "3", "4", "5"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [0, 1, 2, 3, 4],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [10, 10, 10, 10, 10],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [False] * 5,
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
        }
    )

    result = (
        QueryBuilder()
        .register("tbl", dt)
        .execute("select * from tbl order by id asc")
        .read_all()
    )
    last_action = dt.history(1)[0]

    assert last_action["operation"] == "MERGE"
    assert result == expected


@pytest.mark.parametrize("streaming", (True, False))
def test_merge_when_not_matched_by_source_update_with_predicate(
    tmp_path: pathlib.Path, sample_table: Table, streaming: bool
):
    write_deltalake(tmp_path, sample_table, mode="append")

    dt = DeltaTable(tmp_path)

    source_table = Table(
        {
            "id": Array(
                ["6", "7"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [10, 100],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [10, 20],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [False, False],
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
        }
    )

    dt.merge(
        source=source_table,
        source_alias="source",
        target_alias="target",
        predicate="target.id = source.id",
        streamed_exec=streaming,
    ).when_not_matched_by_source_update(
        updates={
            "sold": "int'10'",
        },
        predicate="target.price > 3",
    ).execute()

    expected = Table(
        {
            "id": Array(
                ["1", "2", "3", "4", "5"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [0, 1, 2, 3, 4],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [0, 1, 2, 3, 10],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [False] * 5,
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
        }
    )

    result = (
        QueryBuilder()
        .register("tbl", dt)
        .execute("select * from tbl order by id asc")
        .read_all()
    )
    last_action = dt.history(1)[0]

    assert last_action["operation"] == "MERGE"
    assert result == expected


@pytest.mark.parametrize("streaming", (True, False))
def test_merge_when_not_matched_by_source_delete_with_predicate(
    tmp_path: pathlib.Path, sample_table: Table, streaming: bool
):
    write_deltalake(tmp_path, sample_table, mode="append")

    dt = DeltaTable(tmp_path)

    source_table = Table(
        {
            "id": Array(
                ["6", "7"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [10, 100],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [10, 20],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [False, False],
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
        }
    )

    dt.merge(
        source=source_table,
        source_alias="source",
        target_alias="target",
        predicate="target.id = source.id",
        streamed_exec=streaming,
    ).when_not_matched_by_source_delete(predicate="target.price > 3").execute()

    expected = Table(
        {
            "id": Array(
                ["1", "2", "3", "4"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [0, 1, 2, 3],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [0, 1, 2, 3],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [False] * 4,
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
        }
    )

    result = (
        QueryBuilder()
        .register("tbl", dt)
        .execute("select * from tbl order by id asc")
        .read_all()
    )
    last_action = dt.history(1)[0]

    assert last_action["operation"] == "MERGE"
    assert result == expected


@pytest.mark.parametrize("streaming", (True, False))
def test_merge_when_not_matched_by_source_delete_wo_predicate(
    tmp_path: pathlib.Path, sample_table: Table, streaming: bool
):
    write_deltalake(tmp_path, sample_table, mode="append")

    dt = DeltaTable(tmp_path)

    source_table = Table(
        {
            "id": Array(
                ["4", "5"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "weight": Array(
                [1.5, 1.6],
                ArrowField("weight", type=DataType.float64(), nullable=True),
            ),
        }
    )

    dt.merge(
        source=source_table,
        source_alias="source",
        target_alias="target",
        predicate="target.id = source.id",
        streamed_exec=streaming,
    ).when_not_matched_by_source_delete().execute()

    expected = Table(
        {
            "id": Array(
                ["4", "5"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [3, 4],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [3, 4],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [False] * 2,
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
        }
    )

    result = (
        QueryBuilder()
        .register("tbl", dt)
        .execute("select * from tbl order by id asc")
        .read_all()
    )
    last_action = dt.history(1)[0]

    assert last_action["operation"] == "MERGE"
    assert result == expected


@pytest.mark.parametrize("streaming", (True, False))
def test_merge_multiple_when_matched_update_with_predicate(
    tmp_path: pathlib.Path, sample_table: Table, streaming: bool
):
    write_deltalake(tmp_path, sample_table, mode="append")

    dt = DeltaTable(tmp_path)

    source_table = Table(
        {
            "id": Array(
                ["4", "5"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [10, 100],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [10, 20],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [False, True],
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
        }
    )

    dt.merge(
        source=source_table,
        source_alias="source",
        target_alias="target",
        predicate="target.id = source.id",
        streamed_exec=streaming,
    ).when_matched_update(
        updates={"price": "source.price", "sold": "source.sold"},
        predicate="source.deleted = False",
    ).when_matched_update(
        updates={"price": "source.price", "sold": "source.sold"},
        predicate="source.deleted = True",
    ).execute()

    expected = Table(
        {
            "id": Array(
                ["1", "2", "3", "4", "5"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [0, 1, 2, 10, 100],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [0, 1, 2, 10, 20],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [False] * 5,
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
        }
    )

    result = (
        QueryBuilder()
        .register("tbl", dt)
        .execute("select * from tbl order by id asc")
        .read_all()
    )
    last_action = dt.history(1)[0]

    assert last_action["operation"] == "MERGE"
    assert result == expected


@pytest.mark.parametrize("streaming", (True, False))
def test_merge_multiple_when_matched_update_all_with_predicate(
    tmp_path: pathlib.Path, sample_table: Table, streaming: bool
):
    write_deltalake(tmp_path, sample_table, mode="append")

    dt = DeltaTable(tmp_path)

    source_table = Table(
        {
            "id": Array(
                ["4", "5"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [10, 100],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [10, 20],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [False, True],
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
        }
    )

    dt.merge(
        source=source_table,
        source_alias="source",
        target_alias="target",
        predicate="target.id = source.id",
        streamed_exec=streaming,
    ).when_matched_update_all(
        predicate="source.deleted = False",
    ).when_matched_update_all(
        predicate="source.deleted = True",
    ).execute()

    expected = Table(
        {
            "id": Array(
                ["1", "2", "3", "4", "5"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [0, 1, 2, 10, 100],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [0, 1, 2, 10, 20],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [False, False, False, False, True],
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
        }
    )

    result = (
        QueryBuilder()
        .register("tbl", dt)
        .execute("select * from tbl order by id asc")
        .read_all()
    )
    last_action = dt.history(1)[0]

    assert last_action["operation"] == "MERGE"
    assert result == expected


@pytest.mark.parametrize("streaming", (True, False))
def test_merge_multiple_when_not_matched_insert_with_predicate(
    tmp_path: pathlib.Path, sample_table: Table, streaming: bool
):
    write_deltalake(tmp_path, sample_table, mode="append")

    dt = DeltaTable(tmp_path)

    source_table = Table(
        {
            "id": Array(
                ["6", "9"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [10, 100],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [10, 20],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [False, False],
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
        }
    )

    dt.merge(
        source=source_table,
        source_alias="source",
        target_alias="target",
        predicate="target.id = source.id",
        streamed_exec=streaming,
    ).when_not_matched_insert(
        updates={
            "id": "source.id",
            "price": "source.price",
            "sold": "source.sold",
            "deleted": "False",
        },
        predicate="source.price < 50",
    ).when_not_matched_insert(
        updates={
            "id": "source.id",
            "price": "source.price",
            "sold": "source.sold",
            "deleted": "False",
        },
        predicate="source.price > 50",
    ).execute()

    expected = Table(
        {
            "id": Array(
                ["1", "2", "3", "4", "5", "6", "9"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [0, 1, 2, 3, 4, 10, 100],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [0, 1, 2, 3, 4, 10, 20],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [False] * 7,
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
        }
    )

    result = (
        QueryBuilder()
        .register("tbl", dt)
        .execute("select * from tbl order by id asc")
        .read_all()
    )
    last_action = dt.history(1)[0]

    assert last_action["operation"] == "MERGE"
    assert result == expected


@pytest.mark.parametrize("streaming", (True, False))
def test_merge_multiple_when_matched_delete_with_predicate(
    tmp_path: pathlib.Path, sample_table: Table, streaming: bool
):
    write_deltalake(tmp_path, sample_table, mode="append")

    dt = DeltaTable(tmp_path)

    source_table = Table(
        {
            "id": Array(
                ["5", "4"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "weight": Array(
                [1, 2],
                ArrowField("weight", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [1, 2],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [True, False],
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
            "customer": Array(
                ["Adam", "Patrick"],
                ArrowField("customer", type=DataType.string_view(), nullable=True),
            ),
        }
    )

    dt.merge(
        source=source_table,
        predicate="t.id = s.id",
        source_alias="s",
        target_alias="t",
        streamed_exec=streaming,
    ).when_matched_delete("s.deleted = True").when_matched_delete(
        "s.deleted = false"
    ).execute()

    expected = Table(
        {
            "id": Array(
                ["1", "2", "3"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                list(range(3)),
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                list(range(3)),
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [False] * 3,
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
        }
    )

    result = (
        QueryBuilder()
        .register("tbl", dt)
        .execute("select * from tbl order by id asc")
        .read_all()
    )
    last_action = dt.history(1)[0]

    assert last_action["operation"] == "MERGE"
    assert result == expected


@pytest.mark.parametrize("streaming", (True, False))
def test_merge_multiple_when_not_matched_by_source_update_wo_predicate(
    tmp_path: pathlib.Path, sample_table: Table, streaming: bool
):
    """The first match clause that meets the predicate will be executed, so if you do an update
    without an other predicate the first clause will be matched and therefore the other ones are skipped.
    """
    write_deltalake(tmp_path, sample_table, mode="append")

    dt = DeltaTable(tmp_path)

    source_table = Table(
        {
            "id": Array(
                ["6", "7"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [10, 100],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [10, 20],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [False, False],
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
        }
    )

    dt.merge(
        source=source_table,
        source_alias="source",
        target_alias="target",
        predicate="target.id = source.id",
        streamed_exec=streaming,
    ).when_not_matched_by_source_update(
        updates={
            "sold": "int'10'",
        }
    ).when_not_matched_by_source_update(
        updates={
            "sold": "int'100'",
        }
    ).execute()

    expected = Table(
        {
            "id": Array(
                ["1", "2", "3", "4", "5"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [0, 1, 2, 3, 4],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [10, 10, 10, 10, 10],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
            "deleted": Array(
                [False] * 5,
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
        }
    )

    result = (
        QueryBuilder()
        .register("tbl", dt)
        .execute("select * from tbl order by id asc")
        .read_all()
    )
    last_action = dt.history(1)[0]

    assert last_action["operation"] == "MERGE"
    assert result == expected


@pytest.mark.pyarrow
@pytest.mark.parametrize("streaming", (True, False))
def test_merge_date_partitioned_2344(tmp_path: pathlib.Path, streaming: bool):
    from datetime import date

    import pyarrow as pa

    schema = Schema(
        fields=[
            Field("date", PrimitiveType("date")),
            Field("foo", PrimitiveType("string")),
            Field("bar", PrimitiveType("string")),
        ]
    )

    dt = DeltaTable.create(
        tmp_path, schema=schema, partition_by=["date"], mode="overwrite"
    )

    data = pa.table(
        {
            "date": pa.array([date(2022, 2, 1)]),
            "foo": pa.array(["hello"]),
            "bar": pa.array(["world"]),
        }
    )

    expected = data
    dt.merge(
        data,
        predicate="s.date = t.date",
        source_alias="s",
        target_alias="t",
        streamed_exec=streaming,
    ).when_matched_update_all().when_not_matched_insert_all().execute()

    result = dt.to_pyarrow_table()
    last_action = dt.history(1)[0]

    assert last_action["operation"] == "MERGE"
    assert result == expected
    if not streaming:
        assert (
            last_action["operationParameters"].get("predicate")
            == "date = '2022-02-01'::date"
        )
    else:
        # In streaming mode we don't use aggregated stats of the source in the predicate
        assert last_action["operationParameters"].get("predicate") is None


@pytest.mark.pyarrow
@pytest.mark.parametrize(
    "timezone,predicate",
    [
        (
            None,
            "datetime = arrow_cast('2022-02-01T00:00:00.000000', 'Timestamp(Microsecond, None)')",
        ),
        (
            "UTC",
            "datetime = arrow_cast('2022-02-01T00:00:00.000000', 'Timestamp(Microsecond, Some(\"UTC\"))')",
        ),
    ],
)
def test_merge_timestamps_partitioned_2344(tmp_path: pathlib.Path, timezone, predicate):
    from datetime import datetime

    import pyarrow as pa

    schema = Schema(
        fields=[
            Field(
                "datetime",
                PrimitiveType("timestamp")
                if timezone
                else PrimitiveType("timestamp_ntz"),
            ),
            Field("foo", PrimitiveType("string")),
            Field("bar", PrimitiveType("string")),
        ]
    )

    dt = DeltaTable.create(
        tmp_path, schema=schema, partition_by=["datetime"], mode="overwrite"
    )

    data = pa.table(
        {
            "datetime": pa.array(
                [datetime(2022, 2, 1)], pa.timestamp("us", tz=timezone)
            ),
            "foo": pa.array(["hello"]),
            "bar": pa.array(["world"]),
        }
    )

    dt.merge(
        data,
        predicate="s.datetime = t.datetime",
        source_alias="s",
        target_alias="t",
        streamed_exec=False,  # only with streamed execution off can we use stats to create a pruning predicate
    ).when_matched_update_all().when_not_matched_insert_all().execute()

    result = dt.to_pyarrow_table()
    last_action = dt.history(1)[0]

    assert last_action["operation"] == "MERGE"
    assert result == data
    assert last_action["operationParameters"].get("predicate") == predicate


@pytest.mark.parametrize("streaming", (True, False))
def test_merge_stats_columns_stats_provided(tmp_path: pathlib.Path, streaming: bool):
    data = Table(
        {
            "foo": Array(
                ["a", "b", None, None],
                ArrowField("foo", type=DataType.string_view(), nullable=True),
            ),
            "bar": Array(
                [1, 2, 3, None],
                ArrowField("bar", type=DataType.int64(), nullable=True),
            ),
            "baz": Array(
                [1, 1, None, None],
                ArrowField("baz", type=DataType.int64(), nullable=True),
            ),
        }
    )

    write_deltalake(
        tmp_path,
        data,
        mode="append",
        configuration={"delta.dataSkippingStatsColumns": "foo,baz"},
    )
    dt = DeltaTable(tmp_path)
    add_actions_table = dt.get_add_actions(flatten=True)

    def get_value(name: str):
        return add_actions_table.column(name)[0].as_py()

    # x1 has no max, since inf was the highest value
    assert get_value("null_count.foo") == 2
    assert get_value("min.foo") == "a"
    assert get_value("max.foo") == "b"
    assert get_value("null_count.baz") == 2
    assert get_value("min.baz") == 1
    assert get_value("max.baz") == 1

    with pytest.raises(Exception):
        get_value("null_count.bar")

    data = Table(
        {
            "foo": Array(
                ["a"],
                ArrowField("foo", type=DataType.string_view(), nullable=True),
            ),
            "bar": Array(
                [10],
                ArrowField("bar", type=DataType.int64(), nullable=True),
            ),
            "baz": Array(
                [10],
                ArrowField("baz", type=DataType.int64(), nullable=True),
            ),
        }
    )

    dt.merge(
        data,
        predicate="source.foo = target.foo",
        source_alias="source",
        target_alias="target",
        streamed_exec=streaming,
    ).when_matched_update_all().execute()

    dt = DeltaTable(tmp_path)
    add_actions_table = dt.get_add_actions(flatten=True)

    assert dt.version() == 1

    def get_value(name: str):
        return add_actions_table.column(name)[0].as_py()

    # x1 has no max, since inf was the highest value
    assert get_value("null_count.foo") == 2
    assert get_value("min.foo") == "a"
    assert get_value("max.foo") == "b"
    assert get_value("null_count.baz") == 2
    assert get_value("min.baz") == 1
    assert get_value("max.baz") == 10

    with pytest.raises(Exception):
        get_value("null_count.bar")


def test_merge_field_special_characters_delete_2438(tmp_path: pathlib.Path):
    ## See issue: https://github.com/delta-io/delta-rs/issues/2438
    data = Table(
        {
            "x": Array(
                [1, 2, 3],
                ArrowField("x", type=DataType.int64(), nullable=True),
            ),
            "y--1": Array(
                [4, 5, 6],
                ArrowField("y--1", type=DataType.int64(), nullable=True),
            ),
        }
    )

    write_deltalake(tmp_path, data, mode="append")

    dt = DeltaTable(tmp_path)
    new_data = Table(
        {
            "x": Array(
                [2, 3],
                ArrowField("x", type=DataType.int64(), nullable=True),
            ),
        }
    )

    (
        dt.merge(
            source=new_data,
            predicate="target.x = source.x",
            source_alias="source",
            target_alias="target",
        )
        .when_matched_delete()
        .execute()
    )

    expected = Table(
        {
            "x": Array(
                [1],
                ArrowField("x", type=DataType.int64(), nullable=True),
            ),
            "y--1": Array(
                [4],
                ArrowField("y--1", type=DataType.int64(), nullable=True),
            ),
        }
    )

    assert (
        QueryBuilder().register("tbl", dt).execute("select * from tbl").read_all()
    ) == expected


@pytest.mark.parametrize("streaming", (True, False))
def test_merge_preserves_casing_in_quoted_identifiers(
    tmp_path: pathlib.Path, streaming: bool
):
    data = Table(
        {
            "fooBar": Array(
                ["A001", "A002"],
                ArrowField("fooBar", type=DataType.string_view(), nullable=True),
            )
        }
    )

    write_deltalake(tmp_path, data, mode="append")

    dt = DeltaTable(tmp_path)
    source_table = Table(
        {
            "fooBar": Array(
                ["A003"],
                ArrowField("fooBar", type=DataType.string_view(), nullable=True),
            )
        }
    )

    (
        dt.merge(
            source=source_table,
            predicate='target."fooBar" = source."fooBar"',
            source_alias="source",
            target_alias="target",
            streamed_exec=streaming,
        )
        .when_not_matched_insert_all()
        .execute()
    )

    expected = Table(
        {
            "fooBar": Array(
                ["A001", "A002", "A003"],
                ArrowField("fooBar", type=DataType.string_view(), nullable=True),
            )
        }
    )

    result = (
        QueryBuilder()
        .register("tbl", dt)
        .execute('select * from tbl order by "fooBar" asc')
        .read_all()
    )

    last_action = dt.history(1)[0]

    assert last_action["operation"] == "MERGE"
    assert result == expected


@pytest.mark.pandas
@pytest.mark.pyarrow
def test_struct_casting(tmp_path: pathlib.Path):
    import pandas as pd
    import pyarrow as pa

    cols = ["id", "name", "address"]
    data = [
        (
            2,
            "Marry Doe",
            {"street": "123 Main St", "city": "Anytown", "state": "CA"},
        )
    ]
    df = pd.DataFrame(data, columns=cols)
    df_merge = pd.DataFrame(
        [
            (
                2,
                "Merged",
                {"street": "1 Front", "city": "San Francisco", "state": "CA"},
            )
        ],
        columns=cols,
    )
    assert not df.empty

    schema = pa.Table.from_pandas(df=df).schema

    dt = DeltaTable.create(tmp_path, Schema.from_arrow(schema), name="test")
    metadata = dt.metadata()
    assert metadata.name == "test"

    result = (
        dt.merge(
            source=df_merge,
            predicate="t.id = s.id",
            source_alias="s",
            target_alias="t",
        )
        .when_matched_update_all()
        .execute()
    )
    assert result is not None


@pytest.mark.parametrize("streaming", (True, False))
def test_merge_isin_partition_pruning(tmp_path: pathlib.Path, streaming: bool):
    nrows = 5
    data = Table(
        {
            "id": Array(
                [str(x) for x in range(nrows)],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "partition": Array(
                list(range(nrows)),
                ArrowField("partition", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                list(range(nrows)),
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
        }
    )

    write_deltalake(tmp_path, data, mode="append", partition_by="partition")

    dt = DeltaTable(tmp_path)

    source_table = Table(
        {
            "id": Array(
                ["3", "4"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "partition": Array(
                [3, 4],
                ArrowField("partition", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [10, 20],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
        }
    )

    metrics = (
        dt.merge(
            source=source_table,
            predicate="t.id = s.id and t.partition in (3,4)",
            source_alias="s",
            target_alias="t",
            streamed_exec=streaming,
        )
        .when_matched_update_all()
        .execute()
    )

    expected = Table(
        {
            "id": Array(
                ["0", "1", "2", "3", "4"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "partition": Array(
                [0, 1, 2, 3, 4],
                ArrowField("partition", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [0, 1, 2, 10, 20],
                ArrowField("sold", type=DataType.int32(), nullable=True),
            ),
        }
    )

    result = (
        QueryBuilder()
        .register("tbl", dt)
        .execute("select id, partition, sold from tbl order by id asc")
        .read_all()
    )
    last_action = dt.history(1)[0]

    assert last_action["operation"] == "MERGE"
    assert result == expected
    assert metrics["num_target_files_scanned"] == 2
    assert metrics["num_target_files_skipped_during_scan"] == 3


@pytest.mark.pyarrow
@pytest.mark.parametrize("streaming", (True, False))
def test_cdc_merge_planning_union_2908(tmp_path, streaming: bool):
    """https://github.com/delta-io/delta-rs/issues/2908"""
    cdc_path = f"{tmp_path}/_change_data"
    import pyarrow as pa

    data = {
        "id": pa.array([1, 2], pa.int64()),
        "date": pa.array(
            [datetime.date(1970, 1, 1), datetime.date(1970, 1, 2)], pa.date32()
        ),
    }

    table = pa.Table.from_pydict(data)

    dt = DeltaTable.create(
        table_uri=tmp_path,
        schema=Schema.from_arrow(table.schema),
        mode="overwrite",
        partition_by=["id"],
        configuration={
            "delta.enableChangeDataFeed": "true",
        },
    )

    dt.merge(
        source=table,
        predicate="s.id = t.id",
        source_alias="s",
        target_alias="t",
        streamed_exec=streaming,
    ).when_not_matched_insert_all().execute()

    last_action = dt.history(1)[0]

    assert last_action["operation"] == "MERGE"
    assert dt.version() == 1
    assert os.path.exists(cdc_path), "_change_data doesn't exist"


@pytest.mark.pandas
@pytest.mark.pyarrow
def test_merge_non_nullable(tmp_path):
    import re

    import pandas as pd

    from deltalake.schema import Field, PrimitiveType, Schema

    schema = Schema(
        [
            Field("id", PrimitiveType("integer"), nullable=False),
            Field("bool", PrimitiveType("boolean"), nullable=False),
        ]
    )

    dt = DeltaTable.create(tmp_path, schema=schema)
    df = pd.DataFrame(
        columns=["id", "bool"],
        data=[
            [1, True],
            [2, None],
            [3, False],
        ],
    )

    with pytest.raises(
        Exception,
        match=re.escape("Invalid data found:"),
    ):
        dt.merge(
            source=df,
            source_alias="s",
            target_alias="t",
            predicate="s.id = t.id",
        ).when_matched_update_all().when_not_matched_insert_all().execute()


@pytest.mark.pyarrow
def test_merge_when_wrong_but_castable_type_passed_while_merge(
    tmp_path: pathlib.Path, sample_table: Table
):
    import pyarrow as pa
    import pyarrow.parquet as pq

    write_deltalake(tmp_path, sample_table, mode="append")

    dt = DeltaTable(tmp_path)

    source_table = pa.table(
        {
            "id": pa.array(["7", "8"]),
            "price": pa.array(["1", "2"], pa.string()),
            "sold": pa.array([1, 2], pa.int32()),
            "deleted": pa.array([False, False]),
        }
    )
    dt.merge(
        source=source_table,
        predicate="t.id = s.id",
        source_alias="s",
        target_alias="t",
    ).when_not_matched_insert_all().execute()

    table_schema = pq.read_table(
        tmp_path / dt.get_add_actions().column(0)[0].as_py()
    ).schema
    assert table_schema.field("price").type == sample_table["price"].type


@pytest.mark.pyarrow
def test_merge_on_decimal_3033(tmp_path):
    import pyarrow as pa

    data = {
        "timestamp": [datetime.datetime(2024, 3, 20, 12, 30, 0)],
        "altitude": [Decimal("150.5")],
    }

    table = pa.Table.from_pydict(data)

    schema = Schema.from_arrow(
        pa.schema(
            [
                ("timestamp", pa.timestamp("us")),
                ("altitude", pa.decimal128(6, 1)),
            ]
        )
    )

    dt = DeltaTable.create(tmp_path, schema=schema)

    write_deltalake(dt, table, mode="append")

    dt.merge(
        source=table,
        predicate="target.timestamp = source.timestamp",
        source_alias="source",
        target_alias="target",
        streamed_exec=False,
    ).when_matched_update_all().when_not_matched_insert_all().execute()

    dt.merge(
        source=table,
        predicate="target.timestamp = source.timestamp AND target.altitude = source.altitude",
        source_alias="source",
        target_alias="target",
        streamed_exec=False,  # only with streamed execution off can we use stats to create a pruning predicate
    ).when_matched_update_all().when_not_matched_insert_all().execute()

    string_predicate = dt.history(1)[0]["operationParameters"]["predicate"]

    assert (
        string_predicate
        == "timestamp = arrow_cast('2024-03-20T12:30:00.000000', 'Timestamp(Microsecond, None)') AND altitude = '1505'::decimal(4, 1)"
    )


@pytest.mark.polars
@pytest.mark.xfail(reason="polars will require an update")
def test_merge(tmp_path: pathlib.Path):
    import polars as pl
    from polars.testing import assert_frame_equal

    pl.DataFrame({"id": ["a", "b", "c"], "val": [4.0, 5.0, 6.0]}).write_delta(
        tmp_path, mode="overwrite"
    )

    df = pl.DataFrame({"id": ["a", "b", "c", "d"], "val": [4.1, 5, 6.1, 7]})

    df.write_delta(
        tmp_path,
        mode="merge",
        delta_merge_options={
            "predicate": "tgt.id = src.id",
            "source_alias": "src",
            "target_alias": "tgt",
        },
    ).when_matched_update_all().when_not_matched_insert_all().execute()

    new_df = pl.read_delta(str(tmp_path))
    assert_frame_equal(df, new_df, check_row_order=False)



from __future__ import annotations

from collections.abc import Mapping

from deltalake._internal import RawDeltaTable
from deltalake.fs._base_handler import BaseDeltaStorageHandler

try:
    import pyarrow as pa
    from pyarrow.fs import FileSystemHandler

    PYARROW_AVAILABLE = True
except ImportError as e:
    if "pyarrow" in str(e):
        PYARROW_AVAILABLE = False
    else:
        raise


# NOTE  we need to inherit form FileSystemHandler to pass pyarrow's internal type checks.
if PYARROW_AVAILABLE:

    class DeltaStorageHandler(BaseDeltaStorageHandler, FileSystemHandler):
        """
        DeltaStorageHandler is a concrete implementations of a PyArrow FileSystemHandler.
        """

        def open_input_file(self, path: str) -> pa.PythonFile:
            """
            Open an input file for random access reading.

            Args:
                path: The source to open for reading.

            Returns:
                NativeFile
            """
            return pa.PythonFile(self._handler.open_input_file(path))

        def open_input_stream(self, path: str) -> pa.PythonFile:
            """
            Open an input stream for sequential reading.

            Args:
                path: The source to open for reading.

            Returns:
                NativeFile
            """
            return pa.PythonFile(self._handler.open_input_file(path))

        def open_output_stream(
            self, path: str, metadata: dict[str, str] | None = None
        ) -> pa.PythonFile:
            """
            Open an output stream for sequential writing.

            If the target already exists, existing data is truncated.

            Args:
                path: The source to open for writing.
                metadata: If not None, a mapping of string keys to string values.

            Returns:
                NativeFile
            """
            return pa.PythonFile(self._handler.open_output_stream(path, metadata))

        def open_append_stream(self, path: str, metadata: Mapping[str, str]) -> None:
            raise NotImplementedError
else:

    class DeltaStorageHandler(BaseDeltaStorageHandler):  # type: ignore[no-redef]
        """
        DeltaStorageHandler is a concrete implementations of a PyArrow FileSystemHandler.
        """

        def __init__(
            self,
            table_uri: str,
            options: dict[str, str] | None = None,
            known_sizes: dict[str, int] | None = None,
        ) -> None:
            raise ImportError(
                "DeltaStorageHandler requires pyarrow. Please install deltalake[pyarrow] to use this class."
            )

        @classmethod
        def from_table(
            cls,
            table: RawDeltaTable,
            options: dict[str, str] | None = None,
            known_sizes: dict[str, int] | None = None,
        ) -> "DeltaStorageHandler":
            raise ImportError(
                "DeltaStorageHandler requires pyarrow. Please install deltalake[pyarrow] to use this class."
            )



import os
from datetime import date, datetime
from typing import TYPE_CHECKING

import pytest
from arro3.core import Array, DataType, Field, Table

from deltalake import DeltaTable, write_deltalake
from deltalake.exceptions import DeltaError

if TYPE_CHECKING:
    import pyarrow as pa


def _normalize_pyarrow_view_types_for_sort(tbl: "pa.Table") -> "pa.Table":
    """Normalize view types so pyarrow Table.sort_by works.

    PyArrow does not currently support sorting RecordBatches that contain
    `binary_view` (and sometimes `string_view`) columns.
    """

    import pyarrow as pa

    # Older PyArrow versions may not expose view types.
    binary_view = getattr(pa, "binary_view", None)
    string_view = getattr(pa, "string_view", None)
    if binary_view is None and string_view is None:
        return tbl

    fields = []
    columns = []
    for field in tbl.schema:
        col = tbl[field.name]
        ty = field.type

        if binary_view is not None and ty == binary_view():
            ty = pa.binary()
            # PyArrow 16.x lacks casts for view types, so rebuild the array.
            col = pa.array(col.to_pylist(), type=ty)
        elif string_view is not None and ty == string_view():
            ty = pa.string()
            col = pa.array(col.to_pylist(), type=ty)

        fields.append(
            pa.field(field.name, ty, nullable=field.nullable, metadata=field.metadata)
        )
        columns.append(col)

    return pa.Table.from_arrays(
        columns, schema=pa.schema(fields, metadata=tbl.schema.metadata)
    )


def test_read_cdf_partitioned_with_predicate():
    dt = DeltaTable("../crates/test/tests/data/cdf-table/")
    data = dt.load_cdf(0, 3, predicate="birthday = '2023-12-25'").read_all()

    values = list(set(data["birthday"].to_pylist()))
    assert len(values) == 1
    assert values[0] == date(2023, 12, 25)


def test_read_cdf_partitioned():
    dt = DeltaTable("../crates/test/tests/data/cdf-table/")
    b = dt.load_cdf(0, 3).read_all()
    assert sorted(b["id"].to_pylist()) == [
        1,
        2,
        2,
        2,
        3,
        3,
        3,
        4,
        4,
        4,
        5,
        5,
        5,
        6,
        6,
        6,
        7,
        7,
        7,
        7,
        8,
        9,
        10,
    ]
    assert sorted(b["name"].to_pylist()) == [
        "Ada",
        "Bob",
        "Bob",
        "Bob",
        "Borb",
        "Carl",
        "Carl",
        "Carl",
        "Claire",
        "Dave",
        "Dave",
        "Dave",
        "Dennis",
        "Dennis",
        "Dennis",
        "Dennis",
        "Emily",
        "Emily",
        "Emily",
        "Kate",
        "Kate",
        "Kate",
        "Steve",
    ]
    assert sorted(b["_change_type"].to_pylist()) == [
        "delete",
        "insert",
        "insert",
        "insert",
        "insert",
        "insert",
        "insert",
        "insert",
        "insert",
        "insert",
        "insert",
        "update_postimage",
        "update_postimage",
        "update_postimage",
        "update_postimage",
        "update_postimage",
        "update_postimage",
        "update_preimage",
        "update_preimage",
        "update_preimage",
        "update_preimage",
        "update_preimage",
        "update_preimage",
    ]
    assert sorted(b["_commit_version"].to_pylist()) == [
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        2,
        2,
        2,
        2,
        2,
        2,
        3,
    ]
    assert sorted(b["_commit_timestamp"].to_pylist()) == [
        datetime(2023, 12, 22, 17, 10, 18, 828000),
        datetime(2023, 12, 22, 17, 10, 18, 828000),
        datetime(2023, 12, 22, 17, 10, 18, 828000),
        datetime(2023, 12, 22, 17, 10, 18, 828000),
        datetime(2023, 12, 22, 17, 10, 18, 828000),
        datetime(2023, 12, 22, 17, 10, 18, 828000),
        datetime(2023, 12, 22, 17, 10, 18, 828000),
        datetime(2023, 12, 22, 17, 10, 18, 828000),
        datetime(2023, 12, 22, 17, 10, 18, 828000),
        datetime(2023, 12, 22, 17, 10, 18, 828000),
        datetime(2023, 12, 22, 17, 10, 21, 675000),
        datetime(2023, 12, 22, 17, 10, 21, 675000),
        datetime(2023, 12, 22, 17, 10, 21, 675000),
        datetime(2023, 12, 22, 17, 10, 21, 675000),
        datetime(2023, 12, 22, 17, 10, 21, 675000),
        datetime(2023, 12, 22, 17, 10, 21, 675000),
        datetime(2023, 12, 29, 21, 41, 33, 785000),
        datetime(2023, 12, 29, 21, 41, 33, 785000),
        datetime(2023, 12, 29, 21, 41, 33, 785000),
        datetime(2023, 12, 29, 21, 41, 33, 785000),
        datetime(2023, 12, 29, 21, 41, 33, 785000),
        datetime(2023, 12, 29, 21, 41, 33, 785000),
        datetime(2024, 1, 6, 16, 44, 59, 570000),
    ]
    assert sorted(b["birthday"].to_pylist()) == [
        date(2023, 12, 22),
        date(2023, 12, 22),
        date(2023, 12, 22),
        date(2023, 12, 22),
        date(2023, 12, 23),
        date(2023, 12, 23),
        date(2023, 12, 23),
        date(2023, 12, 23),
        date(2023, 12, 23),
        date(2023, 12, 23),
        date(2023, 12, 24),
        date(2023, 12, 24),
        date(2023, 12, 24),
        date(2023, 12, 24),
        date(2023, 12, 24),
        date(2023, 12, 24),
        date(2023, 12, 25),
        date(2023, 12, 25),
        date(2023, 12, 25),
        date(2023, 12, 29),
        date(2023, 12, 29),
        date(2023, 12, 29),
        date(2023, 12, 29),
    ]


def test_read_cdf_non_partitioned():
    dt = DeltaTable("../crates/test/tests/data/cdf-table-non-partitioned/")
    b = dt.load_cdf(0, 3).read_all()

    assert sorted(b["id"].to_pylist()) == [
        1,
        2,
        2,
        2,
        3,
        3,
        3,
        4,
        4,
        4,
        5,
        5,
        5,
        6,
        6,
        6,
        7,
        7,
        7,
        7,
        8,
        9,
        10,
    ]
    assert sorted(b["name"].to_pylist()) == [
        "Ada",
        "Bob",
        "Bob",
        "Bob",
        "Borb",
        "Carl",
        "Carl",
        "Carl",
        "Claire",
        "Dave",
        "Dave",
        "Dave",
        "Dennis",
        "Dennis",
        "Dennis",
        "Dennis",
        "Emily",
        "Emily",
        "Emily",
        "Kate",
        "Kate",
        "Kate",
        "Steve",
    ]
    assert sorted(b["birthday"].to_pylist()) == [
        date(2024, 4, 14),
        date(2024, 4, 14),
        date(2024, 4, 14),
        date(2024, 4, 14),
        date(2024, 4, 14),
        date(2024, 4, 14),
        date(2024, 4, 14),
        date(2024, 4, 14),
        date(2024, 4, 15),
        date(2024, 4, 15),
        date(2024, 4, 15),
        date(2024, 4, 15),
        date(2024, 4, 15),
        date(2024, 4, 15),
        date(2024, 4, 16),
        date(2024, 4, 16),
        date(2024, 4, 16),
        date(2024, 4, 16),
        date(2024, 4, 16),
        date(2024, 4, 16),
        date(2024, 4, 17),
        date(2024, 4, 17),
        date(2024, 4, 17),
    ]
    assert sorted(b["long_field"].to_pylist()) == [
        1,
        1,
        1,
        1,
        2,
        2,
        2,
        3,
        3,
        3,
        4,
        4,
        4,
        5,
        5,
        5,
        6,
        6,
        6,
        6,
        7,
        8,
        99999999999999999,
    ]
    assert sorted(b["boolean_field"].to_pylist()) == [
        True,
        True,
        True,
        True,
        True,
        True,
        True,
        True,
        True,
        True,
        True,
        True,
        True,
        True,
        True,
        True,
        True,
        True,
        True,
        True,
        True,
        True,
        True,
    ]
    assert sorted(b["double_field"].to_pylist()) == [
        3.14,
        3.14,
        3.14,
        3.14,
        3.14,
        3.14,
        3.14,
        3.14,
        3.14,
        3.14,
        3.14,
        3.14,
        3.14,
        3.14,
        3.14,
        3.14,
        3.14,
        3.14,
        3.14,
        3.14,
        3.14,
        3.14,
        3.14,
    ]
    assert sorted(b["smallint_field"].to_pylist()) == [
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
        1,
    ]
    assert sorted(b["_change_type"].to_pylist()) == [
        "delete",
        "insert",
        "insert",
        "insert",
        "insert",
        "insert",
        "insert",
        "insert",
        "insert",
        "insert",
        "insert",
        "update_postimage",
        "update_postimage",
        "update_postimage",
        "update_postimage",
        "update_postimage",
        "update_postimage",
        "update_preimage",
        "update_preimage",
        "update_preimage",
        "update_preimage",
        "update_preimage",
        "update_preimage",
    ]
    assert sorted(b["_commit_version"].to_pylist()) == [
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        0,
        1,
        1,
        1,
        1,
        1,
        1,
        2,
        2,
        2,
        2,
        2,
        2,
        3,
    ]
    assert sorted(b["_commit_timestamp"].to_pylist()) == [
        datetime(2024, 4, 14, 15, 58, 26, 249000),
        datetime(2024, 4, 14, 15, 58, 26, 249000),
        datetime(2024, 4, 14, 15, 58, 26, 249000),
        datetime(2024, 4, 14, 15, 58, 26, 249000),
        datetime(2024, 4, 14, 15, 58, 26, 249000),
        datetime(2024, 4, 14, 15, 58, 26, 249000),
        datetime(2024, 4, 14, 15, 58, 26, 249000),
        datetime(2024, 4, 14, 15, 58, 26, 249000),
        datetime(2024, 4, 14, 15, 58, 26, 249000),
        datetime(2024, 4, 14, 15, 58, 26, 249000),
        datetime(2024, 4, 14, 15, 58, 29, 393000),
        datetime(2024, 4, 14, 15, 58, 29, 393000),
        datetime(2024, 4, 14, 15, 58, 29, 393000),
        datetime(2024, 4, 14, 15, 58, 29, 393000),
        datetime(2024, 4, 14, 15, 58, 29, 393000),
        datetime(2024, 4, 14, 15, 58, 29, 393000),
        datetime(2024, 4, 14, 15, 58, 31, 257000),
        datetime(2024, 4, 14, 15, 58, 31, 257000),
        datetime(2024, 4, 14, 15, 58, 31, 257000),
        datetime(2024, 4, 14, 15, 58, 31, 257000),
        datetime(2024, 4, 14, 15, 58, 31, 257000),
        datetime(2024, 4, 14, 15, 58, 31, 257000),
        datetime(2024, 4, 14, 15, 58, 32, 495000),
    ]


def test_read_cdf_partitioned_projection():
    dt = DeltaTable("../crates/test/tests/data/cdf-table/")
    columns = ["id", "_change_type", "_commit_version"]
    assert columns == dt.load_cdf(0, 3, columns=columns).schema.names


@pytest.mark.pyarrow
def test_delete_unpartitioned_cdf(tmp_path, sample_data_pyarrow: "pa.Table"):
    import pyarrow as pa
    import pyarrow.compute as pc
    import pyarrow.dataset as ds
    import pyarrow.parquet as pq

    cdc_path = f"{tmp_path}/_change_data"

    write_deltalake(
        tmp_path,
        sample_data_pyarrow,
        mode="append",
        configuration={"delta.enableChangeDataFeed": "true"},
    )
    dt = DeltaTable(tmp_path)
    dt.delete("int64 > 2")

    expected_data = (
        ds.dataset(sample_data_pyarrow)
        .to_table(filter=(pc.field("int64") > 2))
        .append_column(
            field_=pa.field("_change_type", pa.string(), nullable=False),
            column=[["delete"] * 2],
        )
    )
    cdc_data = pq.read_table(cdc_path)

    assert os.path.exists(cdc_path), "_change_data doesn't exist"
    assert cdc_data.to_pydict() == expected_data.to_pydict()


@pytest.mark.pyarrow
def test_delete_partitioned_cdf(tmp_path, sample_data_pyarrow: "pa.Table"):
    import pyarrow as pa
    import pyarrow.compute as pc
    import pyarrow.dataset as ds

    write_deltalake(
        tmp_path,
        sample_data_pyarrow,
        mode="overwrite",
        partition_by=["utf8"],
        configuration={"delta.enableChangeDataFeed": "true"},
    )
    dt = DeltaTable(tmp_path)
    dt.delete("int64 > 2")

    expected_data = (
        ds.dataset(sample_data_pyarrow)
        .to_table(filter=(pc.field("int64") > 2))
        .append_column(
            field_=pa.field("_change_type", pa.string(), nullable=False),
            column=[["delete"] * 2],
        )
        .select(["int64", "_change_type"])
        .sort_by("int64")
    )

    # delta-rs returns Arrow view types (e.g. `string_view`) for zero-copy.
    # PyArrow 16.x lacks compute kernels for view types, and even `Table.filter`
    # can fail because it applies `take` across *all* columns.
    # Convert the only columns we need to kernel-backed types.
    raw_cdc_table = pa.table(dt.load_cdf().read_all())
    change_type_py = raw_cdc_table.column(
        raw_cdc_table.schema.get_field_index("_change_type")
    ).to_pylist()
    cdc_table = pa.table(
        {
            "int64": raw_cdc_table.column(
                raw_cdc_table.schema.get_field_index("int64")
            ),
            "_change_type": pa.array(change_type_py, type=pa.string()),
        }
    )

    delete_mask = pa.array([(v == "delete") for v in change_type_py], type=pa.bool_())
    cdc_data = cdc_table.filter(delete_mask).sort_by("int64")

    assert cdc_data.to_pydict() == expected_data.to_pydict()


@pytest.mark.pyarrow
def test_write_predicate_unpartitioned_cdf(tmp_path, sample_data_pyarrow: "pa.Table"):
    import pyarrow as pa
    import pyarrow.compute as pc
    import pyarrow.dataset as ds
    import pyarrow.parquet as pq

    cdc_path = f"{tmp_path}/_change_data"

    write_deltalake(
        tmp_path,
        sample_data_pyarrow,
        mode="append",
        configuration={"delta.enableChangeDataFeed": "true"},
    )

    dt = DeltaTable(tmp_path)
    write_deltalake(
        dt,
        data=ds.dataset(sample_data_pyarrow).to_table(filter=(pc.field("int64") > 2)),
        mode="overwrite",
        predicate="int64 > 2",
        configuration={"delta.enableChangeDataFeed": "true"},
    )

    expected_data = pa.concat_tables(
        [
            ds.dataset(sample_data_pyarrow)
            .to_table(filter=(pc.field("int64") > 2))
            .append_column(
                field_=pa.field("_change_type", pa.string(), nullable=False),
                column=[["delete"] * 2],
            ),
            ds.dataset(sample_data_pyarrow)
            .to_table(filter=(pc.field("int64") > 2))
            .append_column(
                field_=pa.field("_change_type", pa.string(), nullable=False),
                column=[["insert"] * 2],
            ),
        ]
    )
    cdc_data = pq.read_table(cdc_path)

    assert os.path.exists(cdc_path), "_change_data doesn't exist"
    assert cdc_data.sort_by([("_change_type", "ascending")]) == expected_data.sort_by(
        [("_change_type", "ascending")]
    )
    assert dt.to_pyarrow_table().sort_by([("utf8", "ascending")]) == sample_data_pyarrow


@pytest.mark.pyarrow
def test_write_predicate_partitioned_cdf(tmp_path, sample_data_pyarrow: "pa.Table"):
    import pyarrow as pa
    import pyarrow.compute as pc
    import pyarrow.dataset as ds
    import pyarrow.parquet as pq

    cdc_path = f"{tmp_path}/_change_data"

    write_deltalake(
        tmp_path,
        sample_data_pyarrow,
        mode="overwrite",
        partition_by=["utf8"],
        configuration={"delta.enableChangeDataFeed": "true"},
    )
    dt = DeltaTable(tmp_path)
    write_deltalake(
        dt,
        data=ds.dataset(sample_data_pyarrow).to_table(filter=(pc.field("int64") > 3)),
        mode="overwrite",
        predicate="int64 > 3",
        configuration={"delta.enableChangeDataFeed": "true"},
    )

    expected_data = pa.concat_tables(
        [
            ds.dataset(sample_data_pyarrow)
            .to_table(filter=(pc.field("int64") > 3))
            .append_column(
                field_=pa.field("_change_type", pa.string(), nullable=False),
                column=[["delete"] * 1],
            ),
            ds.dataset(sample_data_pyarrow)
            .to_table(filter=(pc.field("int64") > 3))
            .append_column(
                field_=pa.field("_change_type", pa.string(), nullable=False),
                column=[["insert"] * 1],
            ),
        ]
    )

    table_schema = pa.schema(dt.schema())
    table_schema = table_schema.insert(
        len(table_schema), pa.field("_change_type", pa.string(), nullable=False)
    )
    cdc_data = pq.read_table(cdc_path, schema=table_schema)

    assert os.path.exists(cdc_path), "_change_data doesn't exist"
    assert len(os.listdir(cdc_path)) == 1
    expected_data = expected_data.combine_chunks().sort_by(
        [("_change_type", "ascending")]
    )
    cdc_data = cdc_data.combine_chunks().sort_by([("_change_type", "ascending")])

    assert expected_data == cdc_data

    table = _normalize_pyarrow_view_types_for_sort(dt.to_pyarrow_table())
    assert table.sort_by([("utf8", "ascending")]) == sample_data_pyarrow


@pytest.mark.pyarrow
def test_write_overwrite_unpartitioned_cdf(tmp_path, sample_data_pyarrow: "pa.Table"):
    import pyarrow as pa
    import pyarrow.dataset as ds

    cdc_path = f"{tmp_path}/_change_data"

    write_deltalake(
        tmp_path,
        sample_data_pyarrow,
        mode="append",
        configuration={"delta.enableChangeDataFeed": "true"},
    )

    dt = DeltaTable(tmp_path)
    write_deltalake(
        dt,
        data=ds.dataset(sample_data_pyarrow).to_table(),
        mode="overwrite",
        configuration={"delta.enableChangeDataFeed": "true"},
    )
    sort_values = [("_change_type", "ascending"), ("utf8", "ascending")]
    expected_data = (
        ds.dataset(pa.concat_tables([sample_data_pyarrow] * 3))
        .to_table()
        .append_column(
            field_=pa.field("_change_type", pa.string(), nullable=True),
            column=[["delete"] * 5 + ["insert"] * 10],
        )
    ).sort_by(sort_values)

    assert not os.path.exists(cdc_path), (
        "_change_data shouldn't exist since table was overwritten"
    )

    tbl = dt.load_cdf().read_all()

    select_cols = [
        col
        for col in tbl.column_names
        if col not in ["_commit_version", "_commit_timestamp"]
    ]
    actual = _normalize_pyarrow_view_types_for_sort(pa.table(tbl.select(select_cols)))
    assert actual.sort_by(sort_values) == expected_data

    table = _normalize_pyarrow_view_types_for_sort(dt.to_pyarrow_table())
    assert table.sort_by([("utf8", "ascending")]) == sample_data_pyarrow


@pytest.mark.pyarrow
def test_write_overwrite_partitioned_cdf(tmp_path, sample_data_pyarrow: "pa.Table"):
    import pyarrow as pa
    import pyarrow.compute as pc
    import pyarrow.dataset as ds

    cdc_path = f"{tmp_path}/_change_data"

    write_deltalake(
        tmp_path,
        sample_data_pyarrow,
        mode="append",
        partition_by=["int64"],
        configuration={"delta.enableChangeDataFeed": "true"},
    )

    batch2 = ds.dataset(sample_data_pyarrow).to_table(filter=(pc.field("int64") > 3))

    dt = DeltaTable(tmp_path)
    write_deltalake(
        dt,
        data=batch2,
        mode="overwrite",
        predicate="int64 > 3",
        partition_by=["int64"],
        configuration={"delta.enableChangeDataFeed": "true"},
    )

    table_schema = pa.schema(dt.schema())
    table_schema = table_schema.insert(
        len(table_schema), pa.field("_change_type", pa.string(), nullable=False)
    )

    sort_values = [("_change_type", "ascending"), ("utf8", "ascending")]

    first_batch = sample_data_pyarrow.append_column(
        field_=pa.field("_change_type", pa.string(), nullable=True),
        column=[["insert"] * 5],
    )

    expected_data = pa.concat_tables([batch2] * 2).append_column(
        field_=pa.field("_change_type", pa.string(), nullable=True),
        column=[["delete", "insert"]],
    )

    assert not os.path.exists(cdc_path), (
        "_change_data shouldn't exist since a specific partition was overwritten"
    )

    actual = _normalize_pyarrow_view_types_for_sort(pa.table(dt.load_cdf().read_all()))
    assert actual.drop_columns(["_commit_version", "_commit_timestamp"]).sort_by(
        sort_values
    ).select(expected_data.column_names) == pa.concat_tables(
        [first_batch, expected_data]
    ).sort_by(sort_values)


def test_read_cdf_version_out_of_range():
    dt = DeltaTable("../crates/test/tests/data/cdf-table/")

    with pytest.raises(DeltaError) as e:
        dt.load_cdf(4).read_all()

    assert "invalid table version: 4" in str(e).lower()


def test_read_cdf_version_out_of_range_with_flag():
    dt = DeltaTable("../crates/test/tests/data/cdf-table/")
    b = dt.load_cdf(4, allow_out_of_range=True).read_all()

    assert len(b) == 0


def test_read_timestamp_cdf_out_of_range():
    dt = DeltaTable("../crates/test/tests/data/cdf-table/")
    start = "2033-12-22T17:10:21.675Z"

    with pytest.raises(DeltaError) as e:
        dt.load_cdf(starting_timestamp=start).read_all()

    assert "is greater than latest commit timestamp" in str(e).lower()


def test_read_timestamp_cdf_out_of_range_with_flag():
    dt = DeltaTable("../crates/test/tests/data/cdf-table/")

    start = "2033-12-22T17:10:21.675Z"
    b = dt.load_cdf(starting_timestamp=start, allow_out_of_range=True).read_all()

    assert len(b) == 0


def test_read_cdf_last_version(tmp_path):
    data = Table.from_pydict(
        {"foo": Array([1, 2, 3], type=Field("foo", DataType.int32(), nullable=True))}
    )

    expected = Table.from_pydict(
        {
            "foo": Array([1, 2, 3], type=Field("foo", DataType.int32(), nullable=True)),
            "_change_type": Array(
                ["insert", "insert", "insert"],
                type=Field("_change_type", DataType.string_view(), nullable=True),
            ),
            "_commit_version": Array(
                [0, 0, 0],
                type=Field("_commit_version", DataType.int64(), nullable=True),
            ),
        }
    )

    write_deltalake(
        tmp_path,
        data=data,
        configuration={"delta.enableChangeDataFeed": "true"},
    )

    data = (
        DeltaTable(tmp_path)
        .load_cdf(
            starting_version=0,
            ending_version=0,
            allow_out_of_range=False,
            columns=["foo", "_change_type", "_commit_version"],
        )
        .read_all()
    )

    assert expected == data



import pytest
from arro3.core import Array, DataType, Table

from deltalake import write_deltalake
from deltalake.exceptions import DeltaError


def test_unsafe_cast(tmp_path):
    tbl = Table.from_pydict({"foo": Array([1, 2, 3, 200], DataType.uint8())})

    with pytest.raises(
        DeltaError,
        match="Cast error: Failed to convert into Arrow schema: Cast error: Failed to cast foo from Int8 to UInt8: Can't cast value 200 to type Int8",
    ):
        write_deltalake(tmp_path, tbl)


def test_safe_cast(tmp_path):
    tbl = Table.from_pydict({"foo": Array([1, 2, 3, 4], DataType.uint8())})

    write_deltalake(tmp_path, tbl)



import pytest
from arro3.core import Array, DataType, Field, Schema, Table

from deltalake import DeltaTable, write_deltalake
from deltalake.exceptions import DeltaError


@pytest.fixture()
def sample_table() -> Table:
    nrows = 5
    return Table(
        {
            "id": Array(
                ["1", "2", "3", "4", "5"],
                Field("id", type=DataType.string(), nullable=True),
            ),
            "high price": Array(
                list(range(nrows)),
                Field("high price", type=DataType.int64(), nullable=True),
            ),
        },
    )


def test_not_corrupting_expression(tmp_path):
    data = Table.from_pydict(
        {
            "b": Array([1], DataType.int64()),
            "color_column": Array(["red"], DataType.string()),
        },
    )

    data2 = Table.from_pydict(
        {
            "b": Array([1], DataType.int64()),
            "color_column": Array(["blue"], DataType.string()),
        },
    )

    write_deltalake(
        tmp_path,
        data,
        mode="overwrite",
        partition_by=["color_column"],
        predicate="color_column = 'red'",
    )
    write_deltalake(
        tmp_path,
        data2,
        mode="overwrite",
        partition_by=["color_column"],
        predicate="color_column = 'blue'",
    )


def test_not_corrupting_expression_columns_spaced(tmp_path):
    data = Table.from_pydict(
        {
            "b": Array([1], DataType.int64()),
            "color column": Array(["red"], DataType.string()),
        },
    )

    data2 = Table.from_pydict(
        {
            "b": Array([1], DataType.int64()),
            "color column": Array(["blue"], DataType.string()),
        },
    )

    write_deltalake(
        tmp_path,
        data,
        mode="overwrite",
        # partition_by=["color column"],
        predicate="`color column` = 'red'",
    )
    write_deltalake(
        tmp_path,
        data2,
        mode="overwrite",
        # partition_by=["color column"],
        predicate="`color column` = 'blue'",
    )


# fmt: off

@pytest.mark.parametrize("sql_string", [
    "`high price` >= 0",
    '"high price" >= 0',
    "\"high price\" >= 0"
])
def test_add_constraint(tmp_path, sample_table: Table, sql_string: str):
    write_deltalake(tmp_path, sample_table)

    dt = DeltaTable(tmp_path)

    dt.alter.add_constraint({"check_price": sql_string})

    last_action = dt.history(1)[0]
    assert last_action["operation"] == "ADD CONSTRAINT"
    assert dt.version() == 1
    assert dt.metadata().configuration == {
        "delta.constraints.check_price": '"high price" >= 0'
    }
    assert dt.protocol().min_writer_version == 3

    with pytest.raises(DeltaError):
        # Invalid constraint
        dt.alter.add_constraint({"check_price": '"high price" < 0'})

    with pytest.raises(Exception):
        data = Table(
            {
                "id": Array(["1"], DataType.string()),
                "high price": Array([-1], DataType.int64()),
            },
            schema=Schema(
                fields=[
                    Field("id", type=DataType.string(), nullable=True),
                    Field("high price", type=DataType.int64(), nullable=True),
                ]
            ),
        )

        write_deltalake(tmp_path, data, mode="append")

def test_add_multiple_constraint(tmp_path, sample_table: Table):
    write_deltalake(tmp_path, sample_table)

    dt = DeltaTable(tmp_path)

    dt.alter.add_constraint({"check_price": '"high price" >= 0', "min_price": '"high price" < 5'})

    last_action = dt.history(1)[0]
    assert last_action["operation"] == "ADD CONSTRAINT"
    assert dt.version() == 1
    assert dt.metadata().configuration == {
        "delta.constraints.check_price": '"high price" >= 0',
        "delta.constraints.min_price": '"high price" < 5'

    }
    assert dt.protocol().min_writer_version == 3

    with pytest.raises(Exception):
        data = Table(
            {
                "id": Array(["1"], DataType.string()),
                "high price": Array([5], DataType.int64()),
            },
            schema=Schema(
                fields=[
                    Field("id", type=DataType.string(), nullable=True),
                    Field("high price", type=DataType.int64(), nullable=True),
                ]
            ),
        )

        write_deltalake(tmp_path, data, mode="append")



import pytest

from deltalake import DeltaTable


@pytest.mark.polars
def test_partition_encoding_rfc3896(tmp_path):
    import polars as pl
    from polars.testing import assert_frame_equal

    (start, stop) = (32, 255)
    df = pl.DataFrame(
        {
            "a": list(i for i in range(start, stop)),
            "strings": list("a" + chr(i) for i in range(start, stop)),
        }
    )
    print(f"df:\n{df}")

    # write:
    df.write_delta(tmp_path, delta_write_options={"partition_by": "strings"})

    # read:
    partitioned_tbl = DeltaTable(tmp_path)
    pl_df_partitioned = pl.read_delta(partitioned_tbl).sort("a")

    assert_frame_equal(
        df,
        pl_df_partitioned,
    )



import pathlib

import pytest
from arro3.core import Array, DataType, Table
from arro3.core import Field as ArrowField

from deltalake import CommitProperties, DeltaTable, write_deltalake
from deltalake.query import QueryBuilder


@pytest.fixture()
def sample_table():
    nrows = 5
    return Table(
        {
            "id": Array(
                ["1", "2", "3", "4", "5"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                list(range(nrows)),
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                list(range(nrows)),
                ArrowField("sold", type=DataType.int64(), nullable=True),
            ),
            "price_float": Array(
                list(range(nrows)),
                ArrowField("price_float", type=DataType.float64(), nullable=True),
            ),
            "deleted": Array(
                [False] * nrows,
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
        }
    )


def test_update_with_predicate(tmp_path: pathlib.Path, sample_table: Table):
    write_deltalake(tmp_path, sample_table, mode="append")

    dt = DeltaTable(tmp_path)

    expected = Table(
        {
            "id": Array(
                ["1", "2", "3", "4", "5"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [0, 1, 2, 3, 4],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [0, 1, 2, 3, 4],
                ArrowField("sold", type=DataType.int64(), nullable=True),
            ),
            "price_float": Array(
                [0.0, 1.0, 2.0, 3.0, 4.0],
                ArrowField("price_float", type=DataType.float64(), nullable=True),
            ),
            "deleted": Array(
                [False, False, False, False, True],
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
        }
    )

    commit_properties = CommitProperties(custom_metadata={"userName": "John Doe"})
    dt.update(
        updates={"deleted": "True"},
        predicate="price > 3",
        commit_properties=commit_properties,
    )

    result = QueryBuilder().register("tbl", dt).execute("select * from tbl").read_all()
    last_action = dt.history(1)[0]

    assert last_action["operation"] == "UPDATE"
    assert last_action["userName"] == "John Doe"
    assert result == expected


def test_update_wo_predicate(tmp_path: pathlib.Path, sample_table: Table):
    write_deltalake(tmp_path, sample_table, mode="append")

    dt = DeltaTable(tmp_path)

    expected = Table(
        {
            "id": Array(
                ["1", "2", "3", "4", "5"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [0, 1, 2, 3, 4],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [0, 1, 2, 3, 4],
                ArrowField("sold", type=DataType.int64(), nullable=True),
            ),
            "price_float": Array(
                [0.0, 1.0, 2.0, 3.0, 4.0],
                ArrowField("price_float", type=DataType.float64(), nullable=True),
            ),
            "deleted": Array(
                [True, True, True, True, True],
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
        }
    )

    dt.update(updates={"deleted": "True"})

    result = QueryBuilder().register("tbl", dt).execute("select * from tbl").read_all()
    last_action = dt.history(1)[0]

    assert last_action["operation"] == "UPDATE"
    assert result == expected


def test_update_wrong_types_cast(tmp_path: pathlib.Path, sample_table: Table):
    write_deltalake(tmp_path, sample_table, mode="append")

    dt = DeltaTable(tmp_path)

    with pytest.raises(Exception) as excinfo:
        dt.update(updates={"deleted": "'hello_world'"})

    expected = """Generic DeltaTable error: type_coercion\ncaused by\nError during planning: Failed to coerce then (Utf8) and else (Boolean) to common types in CASE WHEN expression"""
    assert str(excinfo.value) == expected


def test_update_wo_predicate_multiple_updates(
    tmp_path: pathlib.Path, sample_table: Table
):
    write_deltalake(tmp_path, sample_table, mode="append")

    dt = DeltaTable(tmp_path)

    expected = Table(
        {
            "id": Array(
                ["1_1", "2_1", "3_1", "4_1", "5_1"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [0, 1, 2, 3, 4],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [0, 1, 4, 9, 16],
                ArrowField("sold", type=DataType.int64(), nullable=True),
            ),
            "price_float": Array(
                [0.0, 1.0, 2.0, 3.0, 4.0],
                ArrowField("price_float", type=DataType.float64(), nullable=True),
            ),
            "deleted": Array(
                [True, True, True, True, True],
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
        }
    )

    dt.update(
        updates={"deleted": "True", "sold": "sold * price", "id": "concat(id,'_1')"},
        error_on_type_mismatch=False,
    )

    result = QueryBuilder().register("tbl", dt).execute("select * from tbl").read_all()
    last_action = dt.history(1)[0]

    assert last_action["operation"] == "UPDATE"
    assert result == expected


def test_update_with_predicate_and_new_values(
    tmp_path: pathlib.Path, sample_table: Table
):
    write_deltalake(tmp_path, sample_table, mode="append")

    dt = DeltaTable(tmp_path)

    expected = Table(
        {
            "id": Array(
                ["1", "2", "3", "4", "new_id"],
                ArrowField("id", type=DataType.string_view(), nullable=True),
            ),
            "price": Array(
                [0, 1, 2, 3, 4],
                ArrowField("price", type=DataType.int64(), nullable=True),
            ),
            "sold": Array(
                [0, 1, 2, 3, 100],
                ArrowField("sold", type=DataType.int64(), nullable=True),
            ),
            "price_float": Array(
                [0.0, 1.0, 2.0, 3.0, 9999.0],
                ArrowField("price_float", type=DataType.float64(), nullable=True),
            ),
            "deleted": Array(
                [False, False, False, False, True],
                ArrowField("deleted", type=DataType.bool(), nullable=True),
            ),
        }
    )

    dt.update(
        new_values={
            "id": "new_id",
            "deleted": True,
            "sold": 100,
            "price_float": 9999,
            "items_in_bucket": ["item4", "item5", "item6"],
        },
        predicate="price > 3",
    )

    result = QueryBuilder().register("tbl", dt).execute("select * from tbl").read_all()
    last_action = dt.history(1)[0]

    assert last_action["operation"] == "UPDATE"
    assert result == expected


def test_update_no_inputs(tmp_path: pathlib.Path, sample_table: Table):
    write_deltalake(tmp_path, sample_table, mode="append")

    dt = DeltaTable(tmp_path)

    with pytest.raises(Exception) as excinfo:
        dt.update()

    assert (
        str(excinfo.value)
        == "Either updates or new_values need to be passed to update the table."
    )


def test_update_to_many_inputs(tmp_path: pathlib.Path, sample_table: Table):
    write_deltalake(tmp_path, sample_table, mode="append")

    dt = DeltaTable(tmp_path)

    with pytest.raises(Exception) as excinfo:
        dt.update(updates={}, new_values={})

    assert (
        str(excinfo.value)
        == "Passing updates and new_values at same time is not allowed, pick one."
    )


def test_update_with_incorrect_updates_input(
    tmp_path: pathlib.Path, sample_table: Table
):
    write_deltalake(tmp_path, sample_table, mode="append")

    dt = DeltaTable(tmp_path)
    updates = {"col": {}}
    with pytest.raises(Exception) as excinfo:
        dt.update(new_values=updates)

    assert (
        str(excinfo.value)
        == "Invalid datatype provided in new_values, only int, float, bool, list, str or datetime or accepted."
    )


def test_update_stats_columns_stats_provided(tmp_path: pathlib.Path):
    data = Table(
        {
            "foo": Array(
                ["a", "b", None, None],
                ArrowField("foo", type=DataType.string(), nullable=True),
            ),
            "bar": Array(
                [1, 2, 3, None],
                ArrowField("bar", type=DataType.int64(), nullable=True),
            ),
            "baz": Array(
                [1, 1, None, None],
                ArrowField("baz", type=DataType.int64(), nullable=True),
            ),
        }
    )

    write_deltalake(
        tmp_path,
        data,
        mode="append",
        configuration={"delta.dataSkippingStatsColumns": "foo,baz"},
    )
    dt = DeltaTable(tmp_path)
    add_actions_table = dt.get_add_actions(flatten=True)

    def get_value(name: str):
        return add_actions_table.column(name)[0].as_py()

    # x1 has no max, since inf was the highest value
    assert get_value("null_count.foo") == 2
    assert get_value("min.foo") == "a"
    assert get_value("max.foo") == "b"
    assert get_value("null_count.baz") == 2
    assert get_value("min.baz") == 1
    assert get_value("max.baz") == 1

    with pytest.raises(Exception):
        get_value("null_count.bar")

    dt.update({"foo": "'hello world'"})

    dt = DeltaTable(tmp_path)
    add_actions_table = dt.get_add_actions(flatten=True)

    def get_value(name: str):
        return add_actions_table.column(name)[0].as_py()

    # x1 has no max, since inf was the highest value
    assert get_value("null_count.foo") == 0
    assert get_value("min.foo") == "hello world"
    assert get_value("max.foo") == "hello world"
    assert get_value("null_count.baz") == 2
    assert get_value("min.baz") == 1
    assert get_value("max.baz") == 1

    with pytest.raises(Exception):
        get_value("null_count.bar")
